from abc import abstractmethod
from functools import lru_cache
import logging
from typing import Mapping

from llama_cpp import LlamaGrammar
from evoprompt.task import Task
from evoprompt.task.task import DatasetDatum


logger = logging.getLogger(__name__)


class TextClassification(Task):
    def _evaluate_sample(self, prompt: str, datum: DatasetDatum):
        gold_label = self._get_gold_label_for_datum(datum)
        class_mapping = self._get_label_mapping()
        response, usage = self.predict(prompt=prompt, datum=datum)
        response = response.lower()
        if self.use_grammar:
            # model output is from label space
            answer_label = class_mapping[response]
        else:
            answer_label = None
            for label in class_mapping.keys():
                if label in response:
                    answer_label = class_mapping[label]
                    break
            else:
                logger.warning(f"Invalid answer: {response}")
                return "failed", usage

        classification_result = "incorrect" if answer_label != gold_label else "correct"
        return classification_result, usage

    # @lru_cache
    def _get_grammar(self, datum: DatasetDatum, verbose: bool = False):
        return LlamaGrammar.from_string(
            "root ::= ({})".format(
                "|".join(('"' + cons + '"') for cons in self._get_label_mapping())
            ),
            verbose=verbose,
        )

    def _get_prompt_text_for_datum(self, datum: DatasetDatum) -> str:
        return "\nInput: " + '"' + self._get_text_for_datum(datum) + '"'

    @abstractmethod
    def _get_label_mapping(self) -> Mapping:
        pass

    @abstractmethod
    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        pass

    @abstractmethod
    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        pass

    def _aggregate_result(self, results: list[str]) -> float:
        num_correct_results = sum(1 for result in results if result == "correct")
        accuracy = num_correct_results / len(results)
        return accuracy

    @property
    def metric_name(self):
        return "accuracy"
